"""Test the dict_hash function for failure cases."""

import pytest
from dict_hash import dict_hash, ALL_AVAILABLE_HASHES


def test_dict_hash_failure():
    """Test to make sure that the hash raises an error for invalid input."""
    with pytest.raises(ValueError):
        dict_hash(0)

    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        with pytest.raises(ValueError):
            consistent_hash_function(0)



# # Dict Hash
# 
# [![Pypi project](https://badge.fury.io/py/dict-hash.svg)](https://badge.fury.io/py/dict-hash)
# [![License](https://img.shields.io/github/license/LucaCappelletti94/dict_hash)](https://github.com/LucaCappelletti94/dict_hash/blob/master/LICENSE)
# [![Pypi total project downloads](https://pepy.tech/badge/dict-hash)](https://pepy.tech/badge/dict-hash)
# [![Github Actions](https://github.com/LucaCappelletti94/dict_hash/actions/workflows/python.yml/badge.svg)](https://github.com/LucaCappelletti94/dict_hash/actions/)
# 
# Python package to hash dictionaries using both default hash and sha256.
# It comes with full support for hashing Pandas & Polars DataFrame/Series objects, Numba objects and Numpy arrays.
# It supports both objects from Pandas 1.x and 2.x and Numpy 1.x and 2.x.
# 
# Furthermore, the library supports objects that can be recursively hashed.
# 
# As we saw this library being used in the wild mostly to create caching libraries and wrappers,
# we'd like to point you to our library, [Cache decorator](https://github.com/zommiommy/cache_decorator).
# 
# ## Why can't I just use the default hash function?
# 
# In Python, dictionaries just aren't hashable. This is because they are mutable objects, and as such, they cannot be hashed.
# If you were to try and run `hash({})`, you would get a `TypeError` exception.
# 
# ## How do I install this package?
# 
# As usual, just download it using pip:
# 
# ```shell
# pip install dict_hash

# 
# ## Usage examples
# 
# The package offers two functions: `sha256` to generate constant sha256 hashes and `dict_hash`, to generate hashes using the native `hash` function.
# 
# ### Session hash with dict_hash
# 
# Obtain a session hash from the given dictionary.
# 
def test_line_37():
    from dict_hash import dict_hash
    from random_dict import random_dict
    from random import randint
    
    d = random_dict(randint(1, 10), randint(1, 10))
    my_hash = dict_hash(d)

# 
# ### Consistent hashes
# 
# Obtain a consistent hash from the given dictionary. Supported methods include `md5`, `sha256`, `sha1`, `sha224`, `sha384`, `sha512`, `sha3_512`, `shake_128`, `shake_256`, `sha3_384`, `sha3_256`, `sha3_224`, `blake2s`, `blake2b`, as provided from the `hashlib` library.
# 
# For instance, to obtain a sha256 hash from the given dictionary:
# 
def test_line_52():
    from dict_hash import sha256
    from random_dict import random_dict
    from random import randint
    
    d = random_dict(randint(1, 10), randint(1, 10))
    my_hash = sha256(d)

# 
# The methods `shake_128` and `shake_256` expose the length paramater to specify the length of the hash digest.
# 
def test_line_63():
    from dict_hash import shake_128
    from random_dict import random_dict
    from random import randint
    
    d = random_dict(randint(1, 10), randint(1, 10))
    my_hash = shake_128(d, hash_length=16)

# 
# ### Approximated hash
# 
# All of the methods shown offer the `use_approximation` parameter,
# which allows you to switch to a more lightweight hashing procedure
# where supported, for the various supported objects. This procedure
# will randomly subsample the provided objects.
# 
# Currently, we support this parameter for NumPy, Polars, and Pandas objects.
# 
def test_line_81():
    from dict_hash import sha256
    from random_dict import random_dict
    from random import randint
    
    d = random_dict(randint(1, 10), randint(1, 10))
    my_hash = sha256(d)
    
    approximated_hash = sha256(d, use_approximation=True)

# 
# ### Behavior on error
# 
# If the hashing function encounters an object that it cannot hash,
# it will by default raise a `NotHashableException` exception. You
# can choose whether this or other options happen by setting the
# `behavior_on_error` parameter. You can choose between:
# 
# - `raise`: Raise a `NotHashableException` exception.
# - `warn`: Print a `NotHashableWarning` and continue hashing, setting the unhashable object to `"Unhashable object"` string.
# - `ignore`: Ignore the object and continue hashing, setting the unhashable object to `"Unhashable object"` string.
# 
# ### Recursive objects
# 
# In Python it is possible to have recursive objects, such as a dictionary that contains itself.
# When you attempt to hash such an object, the hashing function will raise a `RecursionError` exception,
# which you can customize with the `maximal_recursion` parameter, by default equal to `100`. The
# `RecursionError` is most commonly then handled as a `NotHashableException`, and as such you can
# set the `behavior_on_error` parameter to handle it as you see fit.
# 
# ### Hashable
# 
# When handling complex objects within the dictionaries, you may need to implement
# the class Hashable in that object.
# 
# Here is an example:
# 
def test_line_118():
    from dict_hash import Hashable, sha256
    
    class MyHashable(Hashable):
    
        def __init__(self, a: int):
            self._a = a
            self._time = time()
    
        def consistent_hash(self, use_approximation: bool = False) -> str:
            # The use approximation would be useful when the object is too large,
            # while in this example it may be a bit pointless.
            if use_approximation:
                return sha256({
                    "a": self._a
                }, use_approximation=True)
            return sha256({
                "a": self._a
            })

# 
# ## License
# 
# This software is distributed under the MIT license.



"""Python package for hashing dictionaries and other objects."""

from dict_hash.dict_hash import (
    md5,
    sha256,
    sha1,
    sha224,
    sha384,
    sha512,
    sha3_512,
    shake_128,
    shake_256,
    sha3_384,
    sha3_256,
    sha3_224,
    blake2s,
    blake2b,
    dict_hash,
    NotHashableException,
    NotHashableWarning,
)
from dict_hash.hashable import Hashable
from dict_hash.validate_consistent_hash import validate_consistent_hash

ALL_AVAILABLE_HASHES = [
    md5,
    sha1,
    sha224,
    sha256,
    sha384,
    sha512,
    sha3_224,
    sha3_256,
    sha3_384,
    sha3_512,
    shake_128,
    shake_256,
    blake2s,
    blake2b,
]

__all__ = [
    "md5",
    "sha256",
    "sha1",
    "sha224",
    "sha384",
    "sha512",
    "sha3_512",
    "shake_128",
    "shake_256",
    "sha3_384",
    "sha3_256",
    "sha3_224",
    "blake2s",
    "blake2b",
    "ALL_AVAILABLE_HASHES",
    "dict_hash",
    "Hashable",
    "validate_consistent_hash",
    "NotHashableException",
    "NotHashableWarning",
]



"""Submodule testing hashing of frozendict, frozenset and similar objects."""

from dict_hash import dict_hash


def test_frozenset_hash():
    """Test hashing of frozendict objects."""

    s1 = frozenset(["a", "b", "c"])
    s2 = frozenset(["c", "b", "a"])
    s3 = frozenset(["a", "b", "d"])
    first_dict = {"x": s1, "y": s3}
    second_dict = {"x": s2, "y": s3}
    third_dict = {"x": s3, "y": s1}

    hash1 = dict_hash(first_dict)
    hash2 = dict_hash(second_dict)
    hash3 = dict_hash(third_dict)
    assert hash1 == hash2, "Hashes of identical frozensets should match."
    assert (
        hash1 != hash3
    ), f"Hashes of different frozensets ({first_dict}, {third_dict}) should not match."



"""Module containing the main function of the package."""

import datetime
import hashlib
import inspect
import json
import warnings
from typing import Dict, List, Any, Tuple, Callable
import re
from deflate_dict import deflate

from dict_hash.hashable import Hashable


class NotHashableException(Exception):
    """Exception raised when an object is not hashable."""


class NotHashableWarning(UserWarning):
    """Warning raised when an object is not hashable."""


IGNORED_UNASHABLE_OBJECT_ATTRIBUTES = [
    "_repr_html_",
]


def is_built_in_attribute(obj: Any, attribute_name: str) -> bool:
    """Returns whether attribute is builtin"""
    try:
        attribute = getattr(obj, attribute_name)
        return (attribute.__class__.__module__ in ("__builtin__", "builtins")) or type(
            attribute
        ).__name__ in ("method-wrapper", "builtin_function_or_method")
    except (AttributeError, ValueError):
        return True


def _convert(
    data: Any,
    current_depth: int = 0,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> Any:
    """Returns given data as an hashable object or dictionary.

    Parameters
    ------------------
    data: object
        The data
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Hashable object or dictionary.

    Raises
    ------------------
    NotHashableException
        When we have no clue what to do with the provided object yet and we need to raise an error.

    Warns
    ------------------
    NotHashableWarning
        When we have no clue what to do with the provided object yet and we need to warn the user.
    """
    if current_depth > maximal_recursion:
        raise RecursionError(
            (
                f"Recursion depth exceeded {maximal_recursion}. "
                "Please consider increasing the maximal recursion depth "
                "or simplifying the object to hash."
            )
        )

    # If the object is a None.
    if data is None:
        return "None"
    # If given object is of type Hashable
    if isinstance(data, Hashable):
        # we call its method to convert it to an hash
        # that can be further hashed as required.
        try:
            return data.consistent_hash(use_approximation=use_approximation)
        except TypeError:
            # If this Hashable class is using the legacy interface, we call
            # it without any arguments and raise a warning to advise the user
            # that in a future version this will be deprecated.
            warnings.warn(
                (
                    "The method consistent_hash should take an argument "
                    "use_approximation: bool = False. This is needed to "
                    "allow for the approximation of the hash. "
                    "Please update the method to the new interface. "
                    "This will be deprecated in a future version."
                ),
                DeprecationWarning,
            )
            return data.consistent_hash()
    # If we are handling an instance of fixed-length numpy strings we need to
    # convert them back to a normal python string so that they may be hashed.
    if isinstance(data, bytes):
        return data.decode()
    # If given object is either a date or datetime object
    if isinstance(data, datetime.date):
        # we convert the object to the string version
        # following the ISO format of the date.
        return data.isoformat()

    # converts timedelta to a float
    if isinstance(data, datetime.timedelta):
        return data.total_seconds()

    ############################################
    # Handling hashing of numpy string objects #
    ############################################

    try:
        import numpy as np  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:
        pass
    else:
        # If the given object is a numpy integer, we convert it to a python integer.
        if isinstance(
            data,
            (
                np.uint64,
                np.uint32,
                np.uint16,
                np.uint8,
                np.int64,
                np.int32,
                np.int16,
                np.int8,
            ),
        ):
            return int(data)

        # If the given object is a numpy float, we convert it to a python float.

        try:
            if isinstance(data, (np.float64, np.float32, np.float16, np.float_)):  # type: ignore
                return float(data)
        except AttributeError:
            if isinstance(data, (np.float64, np.float32, np.float16)):
                return float(data)

        # If the given object is a numpy string, we convert it to a python string.
        try:
            string_np_types = (np.str_, np.bytes_)
        except AttributeError:
            string_np_types = (np.str_, np.string_)  # type: ignore

        if isinstance(data, string_np_types):
            return str(data)

    # If the given data is a simple object such as a string, an integer
    # or a float we can leave it to be hashed.
    if isinstance(data, (str, int, float)):
        return data

    ############################################
    # Handling hashing of pandas objects       #
    ############################################

    try:
        import pandas as pd  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:
        pass
    else:
        # A similar behaviour is required for DataFrames.
        if isinstance(data, pd.DataFrame):
            # We store the initial shape of the dataframe, so
            # we can make it part of the hash even if we
            # use the approximation and sample only a part
            # of the dataframe.
            shape = data.shape

            if use_approximation:
                # We take at most the first 50 columns.
                # This is needed because we have encountered DataFrames
                # with millions of columns. Peace to the soul that made them.
                if data.shape[1] > 50:
                    data = data[data.columns[:50]]
                # We sample 50 random lines of the dataframe, as some dataframes
                # can contain millions of samples.
                if data.shape[0] > 50:
                    data = data.sample(n=50, random_state=42)
            return _convert(
                {
                    "hash": _convert(
                        data.to_dict(),
                        current_depth=current_depth + 1,
                        use_approximation=use_approximation,
                        behavior_on_error=behavior_on_error,
                        maximal_recursion=maximal_recursion,
                    ),
                    "shape": shape,
                },
                current_depth=current_depth + 1,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )
        if isinstance(data, pd.Series):
            shape = data.shape

            if use_approximation:
                if data.shape[0] > 50:
                    data = data.sample(n=50, random_state=42)

            return _convert(
                {
                    "hash": _convert(
                        data.to_dict(),
                        current_depth=current_depth + 1,
                        use_approximation=use_approximation,
                        behavior_on_error=behavior_on_error,
                        maximal_recursion=maximal_recursion,
                    ),
                    "name": data.name,
                },
                current_depth=current_depth + 1,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )

    ############################################
    # Handling hashing of Polars objects       #
    ############################################

    try:
        import polars as pl  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:
        pass
    else:
        # A similar behaviour is required for DataFrames.
        if isinstance(data, pl.DataFrame):
            # We store the initial shape of the dataframe, so
            # we can make it part of the hash even if we
            # use the approximation and sample only a part
            # of the dataframe.
            shape = data.shape

            if use_approximation:
                # We take at most the first 50 columns.
                # This is needed because we have encountered DataFrames
                # with millions of columns. Peace to the soul that made them.
                if data.shape[1] > 50:
                    data = data[data.columns[:50]]
                # We sample 50 random lines of the dataframe, as some dataframes
                # can contain millions of samples.
                if data.shape[0] > 50:
                    data = data.sample(n=50)
            return _convert(
                {
                    "hash": _convert(
                        data.to_dict(),
                        current_depth=current_depth + 1,
                        use_approximation=use_approximation,
                        behavior_on_error=behavior_on_error,
                        maximal_recursion=maximal_recursion,
                    ),
                    "shape": shape,
                },
                current_depth=current_depth + 1,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )
        if isinstance(data, pl.Series):
            number_of_elements: int = data.shape[0]

            if use_approximation:
                if number_of_elements > 50:
                    data = data.sample(n=50)

            return _convert(
                {
                    "hash": _convert(
                        np.array(data),
                        current_depth=current_depth + 1,
                        use_approximation=use_approximation,
                        behavior_on_error=behavior_on_error,
                        maximal_recursion=maximal_recursion,
                    ),
                    "name": data.name,
                },
                current_depth=current_depth + 1,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )

    ############################################
    # Handling hashing of numpy array objects  #
    ############################################

    try:
        import numpy as np  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:
        pass
    else:
        # And numpy arrays.
        if isinstance(data, np.ndarray):
            # We store the initial shape of the array, so
            # we can make it part of the hash even if we
            # use the approximation and sample only a part
            # of the array.
            array_shape: Tuple[int, ...] = data.shape

            # We reshape the array so it is always 2D,
            # with at most 50 columns.
            if data.ndim == 1:
                data = data.reshape(-1, 1)
            if data.ndim > 2:
                product_of_dimensions = np.prod(data.shape[1:])
                data = data.reshape(-1, product_of_dimensions)

            if use_approximation:
                if data.shape[1] > 50:
                    data = data[:, :50]

                # We sample 100 random lines of the dataframe, as some dataframes
                # can contain millions of samples.
                if data.shape[0] > 50:
                    pnrg = np.random.RandomState(42)  # pylint: disable=no-member
                    data = data[pnrg.randint(data.shape[0], size=50)]

            return _convert(
                {
                    "hash": _convert(
                        data.tolist(),
                        current_depth=current_depth + 1,
                        use_approximation=use_approximation,
                        behavior_on_error=behavior_on_error,
                        maximal_recursion=maximal_recursion,
                    ),
                    "shape": array_shape,
                },
                current_depth=current_depth + 1,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )

    ############################################
    # Handling hashing of numba array objects  #
    ############################################

    try:
        from numba import typed  # pylint: disable=import-outside-toplevel
    except (ModuleNotFoundError, ImportError):
        pass
    else:
        try:
            # And iterables such as lists and tuples.
            if isinstance(data, typed.List):
                return [
                    _convert(
                        e,
                        current_depth=current_depth + 1,
                        use_approximation=use_approximation,
                        behavior_on_error=behavior_on_error,
                        maximal_recursion=maximal_recursion,
                    )
                    for e in data
                ]
            # If it is a dictionary we need to hash every element of it.
            if isinstance(data, typed.Dict):
                return dict(
                    [
                        _convert(
                            (key, value),
                            current_depth=current_depth + 1,
                            use_approximation=use_approximation,
                            behavior_on_error=behavior_on_error,
                            maximal_recursion=maximal_recursion,
                        )
                        for key, value in data.items()
                    ]
                )
        # In some old numba versions there is no attribute
        # List of Dict in typed.
        except AttributeError:
            pass

    # And iterables such as lists and tuples.
    if isinstance(data, list):
        return [
            _convert(
                e,
                current_depth=current_depth + 1,
                use_approximation=use_approximation,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )
            for e in data
        ]

    # If it is a dictionary we need to hash every element of it.
    if isinstance(data, dict):
        return dict(
            _convert(
                (key, value),
                current_depth=current_depth + 1,
                use_approximation=use_approximation,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )
            for key, value in data.items()
        )

    # And iterables such as lists and tuples.
    if isinstance(data, tuple):
        return tuple(
            _convert(
                e,
                current_depth=current_depth + 1,
                use_approximation=use_approximation,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            )
            for e in data
        )

    if isinstance(data, (set, frozenset)):
        return sorted(
            [
                _convert(
                    e,
                    current_depth=current_depth + 1,
                    use_approximation=use_approximation,
                    behavior_on_error=behavior_on_error,
                    maximal_recursion=maximal_recursion,
                )
                for e in data
            ]
        )

    if isinstance(data, re.Pattern):
        return data.pattern

    if callable(data):
        return "".join(inspect.getsourcelines(data)[0])

    ############################################
    # Handling hashing of Ensmallen objects    #
    ############################################

    try:
        from ensmallen import Graph  # type: ignore
    except ModuleNotFoundError:
        pass
    else:
        if isinstance(data, Graph):
            return data.hash()

    try:
        return _sanitize(
            {
                key: getattr(data, key)
                for key in dir(data)
                if (
                    key not in IGNORED_UNASHABLE_OBJECT_ATTRIBUTES
                    and not is_built_in_attribute(data, key)
                )
            },
        )
    except (NotHashableException, TypeError, RecursionError):
        pass

    if behavior_on_error == "ignore":
        return "Unhashable object"

    message = (
        f"Object of class {data.__class__.__name__} not currently supported. "
        f"The module of origin of the object is {data.__class__.__module__}. "
        "You can easily adding support for the object "
        "by extending the `Hashable` abstract class for "
        "your object of interest. "
        "If you believe this object to be of extremely common use, "
        "please do consider opening up an issue and related "
        "pull requested on the `dict_hash` GitHub repository "
        "to add support for this new type of object. "
        "We only consider for the inclusion in the library "
        "either extremely commonly used objects or objects "
        "that impact often our projects."
    )

    if behavior_on_error == "warn":
        warnings.warn(message, NotHashableWarning)
        return "Unhashable object"

    # Otherwise we need to raise an exception to warn the user.
    raise NotHashableException(message)


def _sanitize(
    dictionary: Dict,
    current_depth: int = 0,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return given dictionary as JSON string.

    Parameters
    -------------------
    dictionary: Dict
        Dictionary to be converted to JSON.
    current_depth: int = 0
        Current recursion depth.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    -------------------
    JSON string representation of given dictionary.

    Raises
    -------------------
    ValueError
        When the given object is not a dictionary.
    ValueError
        When the given value for `behavior_on_error` is not supported.
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".

    Warns
    -------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    if not isinstance(dictionary, (Dict, List)):
        raise ValueError(
            (
                "Given object to hash is not a dictionary nor a List, "
                f"but a {dictionary.__class__.__name__} object, which is not currently supported."
            )
        )
    if not isinstance(behavior_on_error, str) or behavior_on_error not in (
        "raise",
        "warn",
        "ignore",
    ):
        raise ValueError(
            (
                "Given value for `behavior_on_error` is not supported. "
                "Please provide a value in ('raise', 'warn', 'ignore')."
            )
        )
    return json.dumps(
        deflate(
            _convert(
                dictionary,
                current_depth=current_depth,
                use_approximation=use_approximation,
                behavior_on_error=behavior_on_error,
                maximal_recursion=maximal_recursion,
            ),
            leave_tuples=True,
        ),
        sort_keys=True,
    )


def dict_hash(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> int:
    """Return hash of given dict (may not be equal for every session).

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Session hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return hash(
        _sanitize(
            dictionary,
            use_approximation=use_approximation,
            behavior_on_error=behavior_on_error,
            maximal_recursion=maximal_recursion,
        )
    )


def _basic_hash(
    dictionary: Dict,
    hash_function: Callable[[bytes], Any],
    hexdigest_args: Tuple = (),
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    return hash_function(
        _sanitize(
            dictionary,
            use_approximation=use_approximation,
            behavior_on_error=behavior_on_error,
            maximal_recursion=maximal_recursion,
        ).encode("utf-8")
    ).hexdigest(*hexdigest_args)


def md5(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return md5 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.md5,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha1(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha1 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha1,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha224(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha224 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha224,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha256(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha256 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha256,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha384(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha384 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha384,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha512(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha512 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha512,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def blake2b(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return blake2b of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.blake2b,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def blake2s(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return blake2s of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.blake2s,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha3_224(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha3_224 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha3_224,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha3_256(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha3_256 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha3_256,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha3_384(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha3_384 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha3_384,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def sha3_512(
    dictionary: Dict,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return sha3_512 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.sha3_512,
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def shake_128(
    dictionary: Dict,
    hash_length: int = 32,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return shake_128 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    hash_length: int = 32
        Length of the hash to return.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.shake_128,
        hexdigest_args=(hash_length,),
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )


def shake_256(
    dictionary: Dict,
    hash_length: int = 32,
    use_approximation: bool = False,
    behavior_on_error: str = "raise",
    maximal_recursion: int = 100,
) -> str:
    """Return shake_256 of given dict.

    Parameters
    ------------------
    dictionary: Dict
        Dictionary of which determine an unique hash.
    hash_length: int = 32
        Length of the hash to return.
    use_approximation: bool = False
        Whether to employ approximations, such as sampling
        random values in pandas dataframe (using a fixed deterministic
        random seed) or lines in a numpy array. This is mainly
        needed when you need to hash frequently big pandas dataframes
        and you do not care about generating a very precise hash
        but a decent one will do the trick.
    behavior_on_error: str = "raise"
        Whether to raise an error when an unhashable object is found
        or to return a string representation of the object. The options
        are "raise", "warn" and "ignore". If "warn" is selected, a warning
        will be issued using the `warnings` module. If "ignore" is selected,
        the object will be ignored and the hash will be computed without it
        without raising any error or warning.
    maximal_recursion: int = 100
        Maximum recursion depth allowed.

    Returns
    ------------------
    Deterministic hash for the given dictionary.

    Raises
    ------------------
    NotHashableException
        When an object is not hashable and `behavior_on_error` is set to "raise".
    ValueError
        When `behavior_on_error` is not a string or is not in ("raise", "warn", "ignore").

    Warns
    ------------------
    NotHashableWarning
        When an object is not hashable and `behavior_on_error` is set to "warn".
    """
    return _basic_hash(
        dictionary,
        hashlib.shake_256,
        hexdigest_args=(hash_length,),
        use_approximation=use_approximation,
        behavior_on_error=behavior_on_error,
        maximal_recursion=maximal_recursion,
    )



"""Test suite verifying consistency of hashes"""

from pathlib import Path
import numpy as np
from dict_hash import ALL_AVAILABLE_HASHES, dict_hash
from .utils import create_dict


def test_dict_hash():
    """Test to make sure that the hash is created in a consistent way across runs."""
    d = create_dict()
    assert dict_hash(d) == dict_hash(d)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(d) == consistent_hash_function(d)
        Path(consistent_hash_function(d)).touch()


def test_dict_hash_without_approximation_1d():
    """Test to make sure that the hash is stable for 1d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((731,), 1)]
    assert dict_hash(d) == dict_hash(d)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(d) == consistent_hash_function(d)
        Path(consistent_hash_function(d)).touch()


def test_dict_hash_without_approximation_2d():
    """Test to make sure that the hash is stable for 2d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((873, 10), 2)]
    assert dict_hash(d) == dict_hash(d)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(d) == consistent_hash_function(d)
        Path(consistent_hash_function(d)).touch()


def test_dict_hash_without_approximation_3d():
    """Test to make sure that the hash is stable for 3d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((1013, 10, 2), 3)]
    assert dict_hash(d) == dict_hash(d)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(d) == consistent_hash_function(d)
        Path(consistent_hash_function(d)).touch()


def test_dict_hash_without_approximation_4d():
    """Test to make sure that the hash is stable for 4d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 3, 2), 4)]
    assert dict_hash(d) == dict_hash(d)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(d) == consistent_hash_function(d)
        Path(consistent_hash_function(d)).touch()


def test_dict_hash_with_approximation_1d():
    """Test to make sure that the approximation works for 2d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100,), 5)]
    assert dict_hash(d, use_approximation=True) == dict_hash(d, use_approximation=True)

    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(
            d, use_approximation=True
        ) == consistent_hash_function(d, use_approximation=True)
        Path(consistent_hash_function(d, use_approximation=True)).touch()


def test_dict_hash_with_approximation_2d():
    """Test to make sure that the approximation works for 2d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10), 6)]
    assert dict_hash(d, use_approximation=True) == dict_hash(d, use_approximation=True)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(
            d, use_approximation=True
        ) == consistent_hash_function(d, use_approximation=True)
        Path(consistent_hash_function(d, use_approximation=True)).touch()


def test_dict_hash_with_approximation_3d():
    """Test to make sure that the approximation works for 3d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 2), 7)]
    assert dict_hash(d, use_approximation=True) == dict_hash(d, use_approximation=True)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(
            d, use_approximation=True
        ) == consistent_hash_function(d, use_approximation=True)
        Path(consistent_hash_function(d, use_approximation=True)).touch()


def test_dict_hash_with_approximation_4d():
    """Test to make sure that the approximation works for 4d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 3, 2), 8)]
    assert dict_hash(d, use_approximation=True) == dict_hash(d, use_approximation=True)
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(
            d, use_approximation=True
        ) == consistent_hash_function(d, use_approximation=True)
        Path(consistent_hash_function(d, use_approximation=True)).touch()


def test_dict_hash_with_approximation_4d_with_different_shape():
    """Test to make sure that the approximation works for 4d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 3, 3), 8)]
    assert dict_hash(d, use_approximation=True) == dict_hash(d, use_approximation=True)

    previous = create_dict()
    previous["this_is_big"] = [np.full((100, 10, 3, 2), 8)]

    assert dict_hash(d, use_approximation=True) != dict_hash(
        previous, use_approximation=True
    )

    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        assert consistent_hash_function(
            d, use_approximation=True
        ) == consistent_hash_function(d, use_approximation=True)
        Path(consistent_hash_function(d, use_approximation=True)).touch()



"""Test that the hash of a dictionary is the same regardless of the order of the keys."""

from dict_hash import ALL_AVAILABLE_HASHES, dict_hash


def test_with_different_keys_order():
    """Test that the hash of a dictionary is the same regardless of the order of the keys."""
    d1 = {"tune_best_model": True, "target": "def"}

    d2 = {"target": "def", "tune_best_model": True}

    for approximation in [True, False]:
        assert dict_hash(d1, use_approximation=approximation) == dict_hash(
            d2, use_approximation=approximation
        )
        for consistent_hash_function in ALL_AVAILABLE_HASHES:
            assert consistent_hash_function(
                d1, use_approximation=approximation
            ) == consistent_hash_function(d2, use_approximation=approximation)



"""Tests for the package version."""

from validate_version_code import validate_version_code
from dict_hash.__version__ import __version__


def test_version():
    """Test the package version."""
    assert validate_version_code(__version__)



"""Test the dict_hash function."""

import os
import numpy as np
from dict_hash import ALL_AVAILABLE_HASHES
from .utils import create_dict


def test_dict_hash():
    """Test to make sure that the hash is created."""
    d = create_dict()
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_without_approximation_1d():
    """Test to make sure that the hash is stable for 1d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((731,), 1)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_without_approximation_2d():
    """Test to make sure that the hash is stable for 2d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((873, 10), 2)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_without_approximation_3d():
    """Test to make sure that the hash is stable for 3d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((1013, 10, 2), 3)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_without_approximation_4d():
    """Test to make sure that the hash is stable for 4d arrays."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 3, 2), 4)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_with_approximation_1d():
    """Test to make sure that the hash is stable for 1d arrays with approximations."""
    d = create_dict()
    d["this_is_big"] = [np.full((100,), 5)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d, use_approximation=True)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_with_approximation_2d():
    """Test to make sure that the hash is stable for 2d arrays with approximations."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10), 6)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d, use_approximation=True)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_with_approximation_3d():
    """Test to make sure that the hash is stable for 3d arrays with approximations."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 2), 7)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d, use_approximation=True)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_with_approximation_4d():
    """Test to make sure that the hash is stable for 4d arrays with approximations."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 3, 2), 8)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d, use_approximation=True)
        assert os.path.exists(path)
        os.remove(path)


def test_dict_hash_with_approximation_4d_with_different_shape():
    """Test to make sure that the hash is stable for 4d arrays with approximations."""
    d = create_dict()
    d["this_is_big"] = [np.full((100, 10, 3, 3), 8)]
    for consistent_hash_function in ALL_AVAILABLE_HASHES:
        path = consistent_hash_function(d, use_approximation=True)
        assert os.path.exists(path)
        os.remove(path)



"""Tests for the Hashable class."""

from time import time
import pytest
from dict_hash import Hashable, validate_consistent_hash, sha256


class MyHashable(Hashable):
    """A class that implements the consistent_hash method."""

    def __init__(self, a: int):
        self._a = a
        self._time = time()

    def consistent_hash(self, use_approximation: bool = False) -> str:
        if use_approximation:
            return sha256({"a": self._a}, use_approximation=True)
        return sha256({"a": self._a})


class MyDeprecatedHashable(Hashable):
    """A class that implements the consistent_hash method."""

    def __init__(self, a: int):
        self._a = a
        self._time = time()

    def consistent_hash(self) -> str:
        return sha256({"a": self._a})


def test_hashable():
    """Test the Hashable class."""
    with pytest.raises((NotImplementedError, TypeError)):
        Hashable().consistent_hash()

    a = MyHashable(2)
    b = MyHashable(2)
    c = MyHashable(3)
    assert validate_consistent_hash(a, b)
    assert validate_consistent_hash(b, a)
    assert not validate_consistent_hash(a, c)
    assert not validate_consistent_hash(b, c)

    assert sha256({"my_hashable": a}) == sha256({"my_hashable": b})


def test_deprecated_hashable():
    """Test that the deprecated Hashable gets the appropriate warning."""
    with pytest.warns(DeprecationWarning):
        a = MyDeprecatedHashable(2)
        b = MyDeprecatedHashable(2)
        c = MyDeprecatedHashable(3)
        assert validate_consistent_hash(a, b)
        assert validate_consistent_hash(b, a)
        assert not validate_consistent_hash(a, c)
        assert not validate_consistent_hash(b, c)

        assert sha256({"my_hashable": a}) == sha256({"my_hashable": b})
